import numpy as np
from binance.enums import SIDE_BUY, SIDE_SELL
from torch.utils.tensorboard import SummaryWriter

from tradeX.bots.binance_bot import BinanceBot
from tradeX.models.EModel import EModel


class SimpleBot(BinanceBot):
    """
    Simple Futures trading bot. In every new interval:
    1. Cancel all orders
    2. Close all positions
    3. Make prediction (high, low, close) with AI model
    4. Long/short and set take profit at a threshold
    """

    class Intent:
        def __init__(self, low, high):
            self.low = low
            self.high = high

    def __init__(self, model: EModel, logger: SummaryWriter, *args, **kargs, ):
        super(SimpleBot, self).__init__(*args, **kargs)
        self.model = model
        self.logger = logger
        self.iter = 0
        self.intent: SimpleBot.Intent = None

    def on_start(self):
        self.logger.add_scalar("USD", self.balance_usd(fetch_from_cache=False), self.iter)

    def on_new_interval(self, pdata):
        # cancel all orders
        self.cancel_all_orders()
        self.cancel_all_positions()
        self.intent = None

        # log GT
        o, h, l, c = float(pdata[-1][1]), float(pdata[-1][2]), float(pdata[-1][3]), float(pdata[-1][4])
        self.logger.add_scalars("GT", {"High": h, "Low": l, "Close": c}, self.iter)

        # log USD
        self.logger.add_scalar("USD", self.balance_usd(fetch_from_cache=False), self.iter)
        if self.balance_usd(fetch_from_cache=True) < 10:
            exit(0)
        # predict and place order next round
        ph, pl, pc = self.model.predict(pdata)
        ph, pl, pc = round(ph), round(pl), round(pc)

        profit = self.estimate_profit_limit_then_limit(SIDE_BUY, pl, self.balance_usd(fetch_from_cache=True), 0.5,
                                                       exit_price=ph,
                                                       leverage=10)
        if profit > 0:
            self.intent = SimpleBot.Intent(low=pl, high=ph)
            print("Intent", pl, ph)

        self.logger.add_scalars("Pred", {"High": ph, "Low": pl, "Close": pc}, self.iter + 1)
        self.iter += 1

    def on_tik(self, realtime_data):
        data = np.array(realtime_data).astype(float).tolist()
        _, o, h, l, c, *_ = data  # c = current price
        if self.intent is not None:
            if self.position is None:  # this mean the intent is not started yet
                if c - self.intent.low < -1:
                    self.long_market(percent_asset=0.5, leverage=10)
                elif self.intent.high - c < -1:
                    self.short_market(percent_asset=0.5, leverage=10)
            else:  # this means there is a position, time to take profit
                if self.position.s == SIDE_BUY and self.intent.high - c < 0:
                    self.stop_market(SIDE_SELL, self.position.q)
                    self.intent = None
                elif self.position.s == SIDE_SELL and c - self.intent.low < 0:
                    self.stop_market(SIDE_BUY, self.position.q)
                    self.intent = None



import argparse


def make_parser():
    parser = argparse.ArgumentParser("Base parser for tradeX experiments")
    parser.add_argument("--train_data", default="data/BTCUSDT_1m_train.json", type=str,
                        help="Path to train data file")
    parser.add_argument("--val_data", default="data/BTCUSDT_1m_val.json", type=str,
                        help="Path to val data file")
    parser.add_argument("--exp", default="default", help="name of the experiment")
    parser.add_argument("--epochs", type=int, default=1000, help="max epochs to train")
    parser.add_argument("--lr", type=float, default=0.01, help="learning rate, default 0.01")
    parser.add_argument("--bsize", type=int, default=8, help="batch size, default 8")

    parser.add_argument("--norm", choices=["True", "False"],
                        help="norm input data or not (gaussian norm)")
    parser.add_argument("-lb", "--look_back", type=int, default=1000, help="length of input sequence (look back)")
    parser.add_argument("-la", "--look_ahead", type=int, default=5,
                        help="length of data sequence a head to examine (look ahead)")
    parser.add_argument("-g", "--gap", type=float, default=40, help="gap to determine bumb/dump")

    parser.add_argument("--ckpt_dir", default="weights", help="Directory to save checkpoint")
    parser.add_argument("--log_dir", default="logs", help="Directory to log training.")
    parser.add_argument("--num_gpus", type=int, default=1)
    parser.add_argument("--num_workers", type=int, default=4)
    return parser



import mlflow
from pytorch_lightning import LightningModule
from torch.nn import LSTM, Linear
from torch.nn.functional import l1_loss
from torch import tanh


class FCLSTM_RegressionModel(LightningModule):
    def __init__(self, d_model: int = 512, output_d=3, input_d=9):
        super(FCLSTM_RegressionModel, self).__init__()

        # modules
        self.fc_in_en = Linear(input_d, d_model)
        self.lstm = LSTM(input_size=d_model, hidden_size=d_model, batch_first=True, dropout=0.1)

        self.fc_reg = Linear(d_model, output_d)
        self.init_weights()

    def init_weights(self) -> None:
        initrange = 0.1
        for layer in [self.fc_in_en, self.fc_reg]:
            layer.weight.data.uniform_(-initrange, initrange)
            layer.bias.data.zero_()
            layer.weight.data.uniform_(-initrange, initrange)

    def forward(self, src):
        """
        :param src: B x L x D_in
        :param tgt: B x W x D_out
        :param tgt_y: same as $tgt
        :return: shifted right of $tgt in shape of B x W x D_out
        """
        x = self.fc_in_en(src)
        x = tanh(x)
        o, (h, c) = self.lstm(x)
        x = h[0]
        x = self.fc_reg(x)
        return x

    def train_val_step(self, batch, batch_id, key_log):
        input = batch["input"].float()  # B x L_back x D_in
        gt_reg = batch["log_return"].float()  # B x L_ahead x D_in
        log_return_pred = self(input)  # B x 3

        loss = l1_loss(log_return_pred, gt_reg, reduction="mean")

        self.log(f"{key_log}_loss", loss.item())
        mlflow.log_metric(f"{key_log}_loss", loss.item())
        return loss

    def training_step(self, batch, batch_id):
        self.log("learning_rate", self.lr_scheduler.get_last_lr()[0])
        return self.train_val_step(batch, batch_id, "train")

    def validation_step(self, batch, batch_id):
        return self.train_val_step(batch, batch_id, "val")

    def set_optimizer(self, optimizer, lr_scheduler):
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler

    def configure_optimizers(self):
        return [self.optimizer], [self.lr_scheduler]



from binance.enums import SIDE_BUY, SIDE_SELL
from torch.utils.tensorboard import SummaryWriter

from tradeX.bots.binance_bot import BinanceBot
from tradeX.models.EModel import EModel


class LogOnlyBot(BinanceBot):
    """
       This bot is to measure model performance on live market
       """

    def __init__(self, model: EModel, logger: SummaryWriter, *args, **kargs):
        super(LogOnlyBot, self).__init__(*args, **kargs)
        self.model = model
        self.logger = logger
        self.iter = 0
        self.num_trade = 0
        self.__last_position = {"q": 0, "s": 0, "tp": 0}  # q = quantity , s = side, tp = take profit value
        self.num_win = 0
        self.num_lose = 0
        self.last_pred = {"High": 0, "Low": 0, "Close": 0}
        self.acc = {"High": [0, 0], "Low": [0, 0], "Close": [0, 0]}
        self.mse = {"High": 0, "Low": 0, "Close": 0}

    def summary_past(self, past_data):
        o, h, l, c = float(past_data[-1][1]), float(past_data[-1][2]), float(past_data[-1][3]), float(past_data[-1][4])

        # log GT
        self.logger.add_scalar("High/GT", h, self.iter)
        self.logger.add_scalar("Low/GT", l, self.iter)
        self.logger.add_scalar("Close/GT", c, self.iter)

        # calculate profit and log USD
        profit = 0
        if self.__last_position["s"] == SIDE_BUY:
            profit = self.estimate_profit_market_then_limit(SIDE_BUY, o, self.usd, 0.95, self.__last_position["tp"], 1)
            if h > self.__last_position["tp"]:
                self.num_win += 1
            else:
                self.num_lose += 1
            self.usd += profit
        if self.__last_position["s"] == SIDE_SELL:
            profit = self.estimate_profit_market_then_limit(SIDE_SELL, o, self.usd, 0.95, self.__last_position["tp"], 1)
            if l < self.__last_position["tp"]:
                self.num_win += 1
            else:
                self.num_lose += 1
        self.usd += profit
        self.logger.add_scalar("USD", self.usd, self.iter)

        # reset position
        self.__last_position = {"s": 0, "enter": 0, "q": 0,
                                "tp": 0}  # always close position as new interval if TP or SL was not reached

        self.logger.add_scalar("Trade/num_win", self.num_win, self.iter)
        self.logger.add_scalar("Trade/num_lose", self.num_lose, self.iter)

        # log model performance
        for idx, k in enumerate(self.last_pred):
            if self.last_pred[k] != 0:
                num_preds = sum(self.acc[k]) + 1

                # mse
                self.mse[k] = (self.mse[k] * (num_preds - 1) + abs(
                    self.last_pred[k] - float(past_data[-1][idx + 2]))) / num_preds

                # binary acc
                is_up_gt = float(past_data[-1][idx + 2]) - float(past_data[-2][idx + 2])
                is_up_pred = self.last_pred[k] - float(past_data[-2][idx + 2])
                if is_up_pred * is_up_gt > 0:
                    self.acc[k][0] += 1  # right
                else:
                    self.acc[k][1] += 1  # wrong

                # log
                self.logger.add_scalar(f"{k}/right", self.acc[k][0], sum(self.acc[k]))
                self.logger.add_scalar(f"{k}/wrong", self.acc[k][1], sum(self.acc[k]))
                self.logger.add_scalar(f"{k}/l1", self.mse[k], sum(self.acc[k]))

    def on_start(self):
        print("Start==============================================")

    def on_new_interval(self, pdata):
        # calculate_acc
        self.summary_past(pdata)

        # predict and place order next round
        h, l, c = float(pdata[-1][2]), float(pdata[-1][3]), float(pdata[-1][4])
        ph, pl, pc = self.model.predict(pdata, ahead=1)
        ph, pl, pc = round(ph), round(pl), round(pc)

        profit_l = self.estimate_profit_market_then_limit(SIDE_BUY, c, self.usd, 0.95, limit_price=ph,
                                                          leverage=10)
        profit_s = self.estimate_profit_market_then_limit(SIDE_SELL, c, self.usd, 0.95, limit_price=pl,
                                                          leverage=10)

        if profit_l > profit_s > 0 and ph > h and pl > l:
            self.num_trade += 1
            self.__last_position = {"s": SIDE_BUY, "enter": c, "tp": ph}

        elif profit_s > profit_l > 0 and pl < l and ph < h:
            self.num_trade += 1
            self.__last_position = {"s": SIDE_SELL, "enter": c, "tp": pl}

        self.logger.add_scalar("Trade/profit_if_place", max(profit_s, profit_s), self.iter)
        self.logger.add_scalar("Trade/Num_all", self.num_trade, self.iter)

        # below is logging for model optimization purpose
        self.last_pred["High"] = ph
        self.last_pred["Low"] = pl
        self.last_pred["Close"] = pc

        self.logger.add_scalar("High/Pred", ph, self.iter + 1)
        self.logger.add_scalar("Low/Pred", pl, self.iter + 1)
        self.logger.add_scalar("Close/Pred", pc, self.iter + 1)

        self.iter += 1



import torch

from tradeX.models.RegressionModel import ITTransformer


def get_model(ckpt, dmodel=60, nhead=6, dimff=128, inputd=9, outputd=3, decoder_d=4, position_encode=False,
              device="cuda:0"):
    model = ITTransformer(d_model=dmodel, nhead=nhead, dim_feedforward=dimff, input_d=inputd, output_d=outputd,
                          d_in_decoder=decoder_d,
                          position_encode=position_encode)
    model.load_state_dict(torch.load(ckpt, map_location="cpu")['state_dict'])
    model.eval()
    model.to(device)
    return model



import torch
from pytorch_lightning import LightningModule
from torch.nn import Module, LSTM, Linear, ReLU


class IT_LSTM(LightningModule):
    def __init__(self, input_size, hidden_size, output_size, drop_out=0):
        super(IT_LSTM, self).__init__()
        self.lstm = LSTM(input_size=input_size, hidden_size=hidden_size, batch_first=True, dropout=drop_out)
        self.output_size = output_size
        self.fc = Linear(hidden_size, output_size)

    def set_loss(self, loss_func):
        self.loss_func = loss_func

    def forward(self, input):
        output, (hn, cn) = self.lstm(input)
        hn = torch.squeeze(hn, 0)  # BS,hidden
        out = self.fc(hn)  # BSxout
        return out

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=0.0001)
        return optimizer

    def training_step(self, batch, batch_id):
        input = batch["input"].float()  # B x L x input_size
        gt = batch["gt"].float()  # B x output_size
        output = self(input)  # B x output_size
        loss = self.loss_func(output, gt)
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch, batch_id):
        input = batch["input"].float()  # B x L x input_size
        gt = batch["gt"].float()  # B x output_size
        output = self(input)  # B x output_size
        loss = self.loss_func(output, gt)
        self.log("val_loss", loss)
        return loss



import random

import numpy as np
from tacv.fileUtils import load_json
from torch.utils.data import Dataset

from tradeX.datasets.consts import SELECTED_FEATURE_IDX


class RegressionDataset(Dataset):

    def __init__(self, file: str, look_back: int, look_ahead: int = 1, feature_engineer=None):
        self.look_back = look_back
        self.look_ahead = look_ahead
        self.data = load_json(file)
        self.feature_engineer = feature_engineer

    def __getitem__(self, item):
        inp = self.data[item:item + self.look_back]
        gt = self.data[item + self.look_back:item + self.look_back + self.look_ahead]

        # convert to numpy
        inp = np.array(inp)
        gt = np.array(gt)

        # take 9 features, except close & open time
        feature_idx = SELECTED_FEATURE_IDX
        inp = inp[:, feature_idx]  # L_back x 9
        gt = gt[:, feature_idx]  # L_ahead x 9

        # log_return = np.log(gt[-1, 1:4]) - np.log(inp[-1, 1:4])
        log_return_high = np.log(gt[:, 1].max()) - np.log(inp[-1, 1])
        log_return_low = np.log(gt[:, 2].min()) - np.log(inp[-1, 2])
        log_return_close = np.log(gt[-1, 3]) - np.log(inp[-1, 3])
        log_return = np.array((log_return_high, log_return_low, log_return_close))
        if self.feature_engineer is not None:
            inp = self.feature_engineer(inp)

        p_mean = inp[:, 0:4].mean()
        p_std = inp[:, 0:4].std()
        mean = inp.mean(axis=0)
        std = inp.std(axis=0)
        mean[0:4] = p_mean
        std[0:4] = p_std

        if np.count_nonzero(std == 0) != 0:
            return self.__getitem__(random.randint(0, len(self) - 1))

        inp = (inp - mean) / std
        gt = (gt - mean[:9]) / std[:9]

        return {
            "input": inp,  # Lb x 9,
            "mean": mean,
            "std": std,
            "ahead": gt,  # La x 9
            "log_return": log_return,  # 3
        }

    def __len__(self):
        return len(self.data) - self.look_back - self.look_ahead + 1



'''A wrapper class for scheduled optimizer '''
import numpy as np


class ScheduledOptim():
    '''A simple wrapper class for learning rate scheduling'''

    def __init__(self, optimizer, lr_mul, d_model, n_warmup_steps):
        self._optimizer = optimizer
        self.lr_mul = lr_mul
        self.d_model = d_model
        self.n_warmup_steps = n_warmup_steps
        self.n_steps = 0
        self.lr = 0

    def step_and_update_lr(self):
        "Step with the inner optimizer"
        self._update_learning_rate()
        self._optimizer.step()

    def zero_grad(self):
        "Zero out the gradients with the inner optimizer"
        self._optimizer.zero_grad()

    def _get_lr_scale(self):
        d_model = self.d_model
        n_steps, n_warmup_steps = self.n_steps, self.n_warmup_steps
        return (d_model ** -0.5) * min(n_steps ** (-0.5), n_steps * n_warmup_steps ** (-1.5))

    def _update_learning_rate(self):
        ''' Learning rate scheduling per step '''

        self.n_steps += 1
        lr = self.lr_mul * self._get_lr_scale()

        for param_group in self._optimizer.param_groups:
            param_group['lr'] = lr
        self.lr = lr



import argparse
import os.path

from tradeX.utils.data_getter import get_data_all_the_time


def make_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--symbol", type=str, default="BTCUSDT", help="trading pair, .e.g BTCUSDT")
    parser.add_argument("--dfrom", type=int, default=1567962000000,
                        help="Default = 1567962000000 (September 2019),miliseconds since epoch, start time to get trading data to now")
    parser.add_argument("--interval", type=str, default="3m",
                        help="interval of a trading candle (.e.g 1m, 3m, 1h, 1d, 1w, 1M,...)")
    parser.add_argument("--save_dir", type=str, default="data")
    return parser


if __name__ == "__main__":
    args = make_parser().parse_args()
    if not os.path.exists(args.save_dir):
        os.makedirs(args.save_dir)
    symbol = args.symbol
    dfrom = args.dfrom
    interval = args.interval
    save_path = f"{args.save_dir}/{symbol}_{interval}.json"
    get_data_all_the_time(symbol, dfrom, interval, save_path=save_path)



from argparse import ArgumentParser
from tacv.fileUtils import load_json, save_json
from tqdm import tqdm
import numpy as np


def make_parser():
    parser = ArgumentParser()
    parser.add_argument("-d", "--data", default="data/BTCUSDT_1m.json", type=str, help="data path json")
    return parser


if __name__ == "__main__":
    args = make_parser().parse_args()
    file = args.data
    data = load_json(file)
    for item in tqdm(data):
        for i in range(len(item)):
            item[i] = float(item[i])
    train_ratio = 0.8
    train_data = data[:int(len(data) * train_ratio)]
    val_data = data[int(len(data) * train_ratio):]

    # just one more step to validate data
    assert np.count_nonzero(np.diff(np.array(train_data)[:, 0]) != 60000) == 0
    assert np.count_nonzero(np.diff(np.array(val_data)[:, 0]) != 60000) == 0
    assert train_data[-1][0] < val_data[0][0]

    save_json(file.replace(".json", "_train.json"), train_data)
    save_json(file.replace(".json", "_val.json"), val_data)



import torch.nn
from torch.nn.functional import cross_entropy, binary_cross_entropy_with_logits, l1_loss


def log_return_loss(input, target, last, mean, std):
    """
    :param input: B x 3 Prediction of high, low, close
    :param target: B x 3 GT high, low, close
    :param last: B x 3 = last high, low, close
    :return:
    """
    input = input * std + mean
    last = last * std + mean
    target = target * std + mean
    log_return_gt = torch.log(target / last)
    log_return_pred = torch.log(input / last)
    return l1_loss(log_return_gt, log_return_pred).mean()


def bce_loss(input, target, alpha=4):
    bce = binary_cross_entropy_with_logits(input, target, reduction="none")

    positives = target != 0
    negatives = target == 0
    prob = torch.exp(-bce)

    negative_loss = bce[negatives] * (1 - prob[negatives]) ** alpha
    positive_loss = bce[positives] * (1 - prob[positives]) ** alpha
    loss = negative_loss.sum() + positive_loss.sum()
    return loss.sum()


def focal_loss(input, target, pos_neg_weight=(1, 0), alpha=4):
    """
    :param input:
    :param target:
    :param alpha:
    :return:
    """
    ce_loss = cross_entropy(input, target, reduction="none")

    prob = torch.exp(-ce_loss)

    # positives = target != 0
    # negatives = target == 0

    loss = (1 - prob) ** alpha * ce_loss
    return loss.sum()
    # negative_loss = ce_loss[negatives] * pos_neg_weight[1]
    # return positive_loss.sum() + negative_loss.sum()


def unscaled_l1_loss(input, target, *args, **kwargs):
    """
    :param input: B x 3
    :param target: B x 3
    :param args:
    :param kwargs:
    :return:
    """
    mm = None
    if len(args) > 0:
        mm = args[0]
    if "mm" in kwargs:
        mm = kwargs["mm"]
    assert mm is not None
    min = torch.concat([item["min"].view(-1, 1) for item in mm[:3]], dim=-1)  # B x 3
    max = torch.concat([item["max"].view(-1, 1) for item in mm[:3]], dim=-1)  # B x 3
    if input.shape != target.shape:
        target = target[..., :input.shape[-1]]
    target = target * (max - min) + min
    input = input * (max - min) + min

    if "reduction" in kwargs:
        reduction = kwargs["reduction"]
    else:
        reduction = "sum"
    return torch.nn.functional.l1_loss(input, target, reduction=reduction)


def l1_loss(input, target, *args, **kwargs):
    """
    :param input:  B x 8
    :param target: B x 8
    :return:
    """
    if input.shape != target.shape:
        target = target[..., :input.shape[-1]]
    if "reduction" in kwargs:
        reduction = kwargs["reduction"]
    else:
        reduction = "sum"
    return torch.nn.functional.l1_loss(input, target, reduction=reduction)


def l2_loss(input, target, *args, **kwargs):
    """
    :param input:  B x 8
    :param target: B x 8
    :return:
    """
    if input.shape != target.shape:
        target = target[..., :input.shape[-1]]
    if "reduction" in kwargs:
        reduction = kwargs["reduction"]
    else:
        reduction = "sum"
    return torch.nn.functional.mse_loss(input, target, reduction=reduction)


def orientation_loss(pred_theta, gt_theta):
    """
    pred_theta : B x 2 presenting sin and cos
    gt_theta: B x 1 # in radian
    """

    theta_diff = torch.atan2(pred_theta[:, 0], pred_theta[:, 1])  # B

    return -1 * torch.cos(theta_diff - gt_theta).mean()



import datetime
import logging
import time


def get_readable_time():
    return datetime.datetime.fromtimestamp(time.time()).strftime('%Y.%m.%d_%H.%M.%S')


def mils_to_datetime(mls):
    return datetime.datetime.fromtimestamp(mls / 1000).strftime('%Y.%m.%d_%H.%M.%S')


def interval_to_millisecs(interval_str):
    assert len(interval_str) <= 3, "must be in {1m 3m 5m 15m 30m 1h 2h 4h 6h 8h 12h 1d 3d 1w 1M}"
    unit = interval_str[-1:]
    num = int(interval_str[:-1])
    if unit == "m":
        return 60 * 1000 * num
    if unit == "h":
        return 60 * 60 * 1000 * num
    if unit == "d":
        return 24 * 60 * 60 * 1000 * num
    if unit == "w":
        return 7 * 24 * 60 * 60 * 1000 * num
    if unit == "M":
        return 30 * 7 * 24 * 60 * 60 * 1000 * num
    logging.warning("Invalid interval, use default interval 3m")
    return 60 * 1000 * 3



import mlflow
import torch
from pytorch_lightning import LightningModule
from torch.nn import TransformerEncoder, TransformerEncoderLayer, LayerNorm, Linear
from torch.nn.functional import leaky_relu

from tradeX.utils.losses import orientation_loss


class DirectionModel(LightningModule):
    """
    Predict at this point will coin price bump, dump or keep stable.
    """

    def __init__(self, sequence_length, d_in=9, d_model=64, nhead=8, dim_ff=64):
        super(DirectionModel, self).__init__()
        self.d_model = d_model
        self.in_fc = Linear(d_in, d_model)
        encoder_layer = TransformerEncoderLayer(d_model, nhead, dim_ff, batch_first=True)
        encoder_norm = LayerNorm(d_model)

        self.encoder = TransformerEncoder(encoder_layer, 6, encoder_norm)
        self.fc_theta = Linear(d_model * sequence_length, 2)

    def forward(self, x):
        """
        :param x: B x L x  d_in
        :return:
        """
        x = self.in_fc(x)  # B x L x d_model
        x = self.encoder(x)  # B x L x d_model
        x = x.view(x.shape[0], -1)  # B x E where E = d_model x L
        x = leaky_relu(x)
        sin_cos = self.fc_theta(x)  # B x 2
        sin_cos = torch.nn.functional.normalize(sin_cos, p=2, dim=-1)  # B x 2

        return sin_cos

    def training_step(self, batch, batch_idx):
        mlflow.log_metric("learning_rate", self.lr_scheduler.get_lr()[0])
        inp = batch["input"]  # B x L x D

        gt_theta = batch["theta"]  # B x 1 in radian
        sin_cos = self(inp)

        coss_loss = orientation_loss(sin_cos, gt_theta).mean()
        sin_loss = torch.zeros(1).item()
        gt_sin = torch.sin(gt_theta)
        pred_sin = sin_cos[:, 0]
        pos_sin_loss_idx = torch.logical_and(gt_sin > 0, pred_sin < gt_sin)
        neg_sin_loss_idx = torch.logical_and(gt_sin < 0, pred_sin > gt_sin)
        if torch.count_nonzero(pos_sin_loss_idx) > 0:
            sin_loss += (gt_sin - pred_sin)[pos_sin_loss_idx].mean()
        if torch.count_nonzero(neg_sin_loss_idx) > 0:
            sin_loss += (pred_sin - gt_sin)[neg_sin_loss_idx].mean()

        mlflow.log_metric("train_loss", coss_loss.item())
        mlflow.log_metric("sin_loss", sin_loss.item())
        self.log("train_loss", coss_loss.item())
        return coss_loss + sin_loss

    def validation_step(self, batch, batch_idx):
        inp = batch["input"]  # B x L x D
        gt_theta = batch["theta"]  # B x 1 in radian
        sin_cos = self(inp)

        loss = orientation_loss(sin_cos, gt_theta).mean()
        mlflow.log_metric("val_loss", loss.item())
        self.log("val_loss", loss.item())
        return loss

    def set_optimizer(self, optimizer, lr_scheduler):
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler

    def configure_optimizers(self):
        assert self.optimizer is not None and self.lr_scheduler is not None
        return [self.optimizer], [self.lr_scheduler]



import numpy as np
import talib

from tradeX.datasets.BinanceDataset import BinanceDataset
from tradeX.datasets.consts import SELECTED_FEATURE_IDX
from scipy.signal import argrelextrema


class DirectionDataset(BinanceDataset):
    EVENT_UP = 1
    EVENT_DOWN = 2
    EVENT_STABLE = 0
    R_MEAN = 11.1565
    R_STD = 9.060607471356432

    def __init__(self, look_ahead, file, look_back, norm=True, limit=-1, feature_engineer=None):
        """
        :param look_ahead:
        :param file:
        :param look_back:
        :param gap: to determine if a candle is up or down. If 0 < abs(openPrice-closePrice) < gap --> STABLE
        :param norm:
        :param limit:
        """
        super(DirectionDataset, self).__init__(file, look_back=look_back, look_ahead=look_ahead, norm=norm,
                                               limit=limit)
        self.feature_engineer = feature_engineer

    def calculate_r(self, lb, all):
        return self.look_ahead
        # ma = talib.EMA(all[:, 0:4].mean(axis=1), timeperiod=4)
        # ma = talib.EMA(ma, timeperiod=2)
        # ma = talib.EMA(ma, timeperiod=2)
        # ma = talib.EMA(ma, timeperiod=2)
        # ma = talib.EMA(ma, timeperiod=2)
        # ma = talib.EMA(ma, timeperiod=2)
        # ma = talib.EMA(ma, timeperiod=2)
        #
        # kline_period = 12
        # local_maximums = argrelextrema(ma, np.greater, order=kline_period)[0]
        # local_minimums = argrelextrema(ma, np.less, order=kline_period)[0]
        #
        # after_lb_local_max = np.nonzero(local_maximums >= lb)[0]
        # after_lb_local_min = np.nonzero(local_minimums >= lb)[0]
        #
        # if len(after_lb_local_min) <= 0 and len(after_lb_local_max) <= 0:
        #     return int(DirectionDataset.R_MEAN)
        # if len(after_lb_local_max) <= 0:
        #     next_min = after_lb_local_min[0]
        #     r = local_minimums[next_min] - lb + 1
        #     return r
        # else:
        #     next_max = after_lb_local_max[0]
        #
        # if len(after_lb_local_min) <= 0:
        #     next_max = after_lb_local_max[0]
        #     r = local_maximums[next_max] - lb + 1
        #     return r
        # else:
        #     next_min = after_lb_local_min[0]
        #
        # if local_maximums[next_max] < local_minimums[next_min]:
        #     return local_maximums[next_max] - lb + 1
        # else:
        #     return local_minimums[next_min] - lb + 1

    def calculate_r_and_theta(self, inp, gt):
        all = np.vstack((inp, gt)).astype(float)
        ma = talib.EMA(all[:, :4].mean(axis=1), timeperiod=2)
        ma = talib.EMA(ma, timeperiod=2)
        ma = talib.EMA(ma, timeperiod=2)
        ma = talib.EMA(ma, timeperiod=2)
        ma = talib.EMA(ma, timeperiod=2)

        lb = len(inp)
        r = self.calculate_r(lb, all)

        theta = self.calculate_theta(ma[-self.look_ahead+2:])

        # current_price = all[:, 3][lb - 1]
        # next_price = ma[lb - 1 + r]
        # theta = np.arctan2(next_price - current_price, 1)
        return theta

    def calculate_theta(self, price):
        x = np.linspace(0, 1, len(price))
        y = price

        # Perform linear regression
        X = np.vstack([x, np.ones(len(price))]).T
        theta = np.linalg.pinv(X) @ y

        # Calculate the angle between the line and the x-axis
        angle = np.arctan(theta[0])
        return angle

    def __getitem__(self, item):
        idx = item + self.look_back
        inp = self.data[item:idx]
        gt = self.data[idx:idx + self.look_ahead]

        # take  9 features, except close & open time
        feature_idx = SELECTED_FEATURE_IDX

        inp = np.array(inp, dtype=np.float32)
        inp = inp[:, feature_idx]

        # take  9 features, except close & open time

        gt = np.array(gt, dtype=np.float32)
        gt = gt[:, feature_idx]

        p_mean = inp[:, 0:4].mean()
        p_std = inp[:, 0:4].std()

        mean = inp.mean(axis=0)
        std = inp.std(axis=0)
        mean[0:4] = p_mean
        std[0:4] = p_std

        inp = (inp - mean) / std
        gt = (gt - mean) / std

        theta = self.calculate_r_and_theta(inp, gt)
        return {
            "input": inp,  # L x 9,
            "theta": theta,
            "mean": mean,
            "std": std,
            "gt": gt
        }

    def __len__(self):
        return len(self.data) - self.look_back - self.look_ahead + 1



import mlflow
import torch
from pytorch_lightning import LightningModule
from torch.nn import TransformerEncoder, TransformerEncoderLayer, LayerNorm, Linear
from torch.nn.functional import leaky_relu, binary_cross_entropy_with_logits, cross_entropy


class BinaryClassifier(LightningModule):
    """
    Predict at this point will coin price bump, dump or keep stable.
    """

    def __init__(self, d_in=9, d_out=3, d_model=64, nhead=8, dim_ff=64, look_back=100):
        super(BinaryClassifier, self).__init__()
        self.d_model = d_model
        self.in_fc = Linear(d_in, d_model)
        encoder_layer = TransformerEncoderLayer(d_model, nhead, dim_ff, batch_first=True)
        encoder_norm = LayerNorm(d_model)
        self.lookback = look_back
        self.encoder = TransformerEncoder(encoder_layer, 6, encoder_norm)
        self.fc_out = Linear(d_model * look_back, d_out)

    def forward(self, x):
        """
        :param x: B x L x  d_in
        :return:
        """
        x = self.in_fc(x)  # B x L x d_model
        x = self.encoder(x)  # B x L x d_model
        x = x.view(x.shape[0], -1)  # B x E where E = d_model x L
        x = leaky_relu(x)
        x = self.fc_out(x)  # B x d_out
        return x

    def training_step(self, batch, batch_idx):
        mlflow.log_metric("learning_rate", self.lr_scheduler.get_lr()[0])
        input = batch["input"]  # B x L x D
        event = batch["event"]  # B x 3
        x = self(input)  # B x 3
        loss = binary_cross_entropy_with_logits(x.view(-1), event, reduction="mean")
        mlflow.log_metric("train_loss", loss.item())
        return loss #np.hstack((x.sigmoid().cpu().detach().numpy(),event.detach().cpu().numpy().reshape(-1,1)))

    def validation_step(self, batch, batch_idx):
        input = batch["input"]  # B x L x D
        event = batch["event"]  # B x 3
        x = self(input)  # B x 3
        loss = binary_cross_entropy_with_logits(x.view(-1), event, reduction="mean")
        self.log("val_loss", loss.item())
        mlflow.log_metric("val_loss", loss.item())
        return loss

    def set_optimizer(self, optimizer, lr_scheduler):
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler

    def configure_optimizers(self):
        assert self.optimizer is not None and self.lr_scheduler is not None
        return [self.optimizer], [self.lr_scheduler]



import ast
import asyncio
import time

import numpy as np
from binance import Client
from binance.enums import SIDE_BUY, SIDE_SELL, FUTURE_ORDER_TYPE_MARKET, FUTURE_ORDER_TYPE_STOP_MARKET, \
    FUTURE_ORDER_TYPE_TAKE_PROFIT_MARKET, FUTURE_ORDER_TYPE_TAKE_PROFIT, FUTURE_ORDER_TYPE_LIMIT, TIME_IN_FORCE_GTC
from loguru import logger
from websocket import WebSocketApp

from tradeX.bots.consts import INTERVAL_2_SECONDS, Mode_DryRun, Mode_Live, Mode_Testnet
from tradeX.utils.data_getter import _get_data_chunk, get_data_all_the_time


class BinanceBot:
    class Position:
        def __init__(self, q, s, entry=0):
            self.q = q
            self.s = s
            self.entry = entry

    def __init__(self, symbol, interval, look_back=100, data_delay_allowed=5000, stop_on_err=True, usd=0):
        self.symbol = symbol
        self.interval = interval
        self.look_back = look_back
        self.data_delay_allowed = data_delay_allowed
        self.data = []
        self.stop_on_err = stop_on_err
        self.socket = WebSocketApp(
            f"wss://fstream.binance.com/stream?streams={self.symbol}@kline_{self.interval}",
            on_message=self.on_message,
            on_error=self.on_error)
        self.client: Client
        self.position: BinanceBot.Position = None  # list of Positions objects
        self.__run_mode = Mode_DryRun
        self.usd = usd

    @staticmethod
    def ws_data_to_kline_data(ws_data):
        open_t = ws_data['t']
        open = ws_data['o']
        high = ws_data['h']
        low = ws_data['l']
        close = ws_data['c']
        volume = ws_data['v']
        close_t = ws_data['T']
        qv = ws_data['q']
        n = ws_data['n']
        tbbav = ws_data['V']
        tbqav = ws_data['Q']
        ignore = ws_data['B']
        return [open_t, open, high, low, close, volume, close_t, qv, n, tbbav, tbqav, ignore]

    def estimate_profit_limit_then_limit(self, side, entry_price, current_usd, percent_asset, exit_price, leverage):
        """
        roughly estimate if LONG/SHORT limit then STOP/TAKE_PROFIT limit
        """
        usd = current_usd
        q = round(usd * percent_asset * leverage / entry_price, 3)

        fee1 = 0.02 / 100 * q * entry_price
        fee2 = 0.02 / 100 * q * exit_price

        if side == SIDE_SELL:
            return (1 / exit_price - 1 / entry_price) * q * entry_price * exit_price - fee1 - fee2
        else:
            return (1 / entry_price - 1 / exit_price) * q * entry_price * exit_price - fee1 - fee2

    def estimate_profit_market_then_limit(self, side, mark_price, current_usd, percent_asset, limit_price, leverage):
        """
        roughly estimate if LONG/SHORT market then STOP/TAKE_PROFIT limit
        """
        usd = current_usd
        q = round(usd * percent_asset * leverage / mark_price, 3)

        fee1 = 0.04 / 100 * q * mark_price
        fee2 = 0.02 / 100 * q * limit_price

        if side == SIDE_SELL:
            return (1 / limit_price - 1 / mark_price) * q * mark_price * limit_price - fee1 - fee2
        else:
            return (1 / mark_price - 1 / limit_price) * q * mark_price * limit_price - fee1 - fee2

    def warning_if_there_is_open_position(self, order_type: str):
        if self.position is not None:
            logger.warning(f"Not allow to place {order_type} order due to there is an opening position")

    def long_market(self, percent_asset=0.9, leverage=1):
        """
        :param percent_asset: control how much of USD to place this order, 1 means ALL IN
        :param leverage:
        :return:
        """
        assert self.__run_mode != Mode_DryRun
        self.warning_if_there_is_open_position(order_type="long market")
        try:
            self.client.futures_change_leverage(symbol=self.symbol, leverage=leverage)
            usd = self.balance_usd()
            p = self.market_price()
            q = round(usd / p * percent_asset * leverage, 3)
            res = self.client.futures_create_order(
                symbol=self.symbol,
                side=SIDE_BUY,
                type=FUTURE_ORDER_TYPE_MARKET,
                quantity=q
            )
            self.position = BinanceBot.Position(q=q, s=SIDE_BUY)
            return res
        except Exception as e:
            logger.exception(e)

    def long_limit(self, price, percent_asset=0.9, leverage=1):
        """
        :param percent_asset: control how much of USD to place this order, 1 means ALL IN
        :param leverage:
        :return:
        """
        assert self.__run_mode != Mode_DryRun
        self.warning_if_there_is_open_position(order_type="long limit")
        try:
            self.client.futures_change_leverage(symbol=self.symbol, leverage=leverage)
            usd = self.balance_usd()
            p = self.market_price()
            q = round(usd / p * percent_asset, 3)
            res = self.client.futures_create_order(
                symbol=self.symbol,
                side=SIDE_BUY,
                type=FUTURE_ORDER_TYPE_LIMIT,
                quantity=q,
                timeInForce=TIME_IN_FORCE_GTC,
                price=price
            )
            self.position = BinanceBot.Position(q=q, s=SIDE_BUY)
            return res
        except Exception as e:
            logger.exception(e)

    def short_market(self, percent_asset=0.9, leverage=1):
        """
                :param percent_asset: control how much of USD to place this order, 1 means ALL IN
                :param leverage:
                :return:
                """
        assert self.__run_mode != Mode_DryRun
        self.warning_if_there_is_open_position(order_type="short market")
        try:
            self.client.futures_change_leverage(symbol=self.symbol, leverage=leverage)
            usd = self.balance_usd()
            p = self.market_price()
            q = round(usd / p * percent_asset * leverage, 3)
            res = self.client.futures_create_order(
                symbol=self.symbol,
                side=SIDE_SELL,
                type=FUTURE_ORDER_TYPE_MARKET,
                quantity=q
            )
            self.position = BinanceBot.Position(s=SIDE_SELL, q=q)
            return res
        except Exception as e:
            logger.exception(e)

    def short_limit(self, price, percent_asset=0.9, leverage=1):
        """
                :param percent_asset: control how much of USD to place this order, 1 means ALL IN
                :param leverage:
                :return:
                """
        assert self.__run_mode != Mode_DryRun
        self.warning_if_there_is_open_position(order_type="short limit")
        try:
            self.client.futures_change_leverage(symbol=self.symbol, leverage=leverage)
            usd = self.balance_usd()
            p = self.market_price()
            q = round(usd / p * percent_asset, 3)
            res = self.client.futures_create_order(
                symbol=self.symbol,
                side=SIDE_SELL,
                type=FUTURE_ORDER_TYPE_LIMIT,
                quantity=q,
                price=price,
                timeInForce=TIME_IN_FORCE_GTC,
            )
            self.position = BinanceBot.Position(s=SIDE_SELL, q=q)
            return res
        except Exception as e:
            logger.exception(e)

    def stop_market(self, side, q):
        assert self.__run_mode != Mode_DryRun
        try:
            self.client.futures_create_order(symbol=self.symbol, side=side,
                                             type=FUTURE_ORDER_TYPE_MARKET, quantity=q, closePosition=True)
        except Exception as e:
            logger.exception(e)

    def take_profit_market(self, stop_price, side):
        assert self.__run_mode != Mode_DryRun
        try:
            self.client.futures_create_order(symbol=self.symbol, stopPrice=stop_price, side=side,
                                             type=FUTURE_ORDER_TYPE_TAKE_PROFIT_MARKET, closePosition=True)
        except Exception as e:
            logger.exception(e)

    def take_profit_limit(self, side, stop_price, price):
        assert self.__run_mode != Mode_DryRun
        try:
            self.client.futures_create_order(symbol=self.symbol, stopPrice=stop_price, price=price, side=side,
                                             type=FUTURE_ORDER_TYPE_TAKE_PROFIT, closePosition=True)
        except Exception as e:
            logger.exception(e)

    def cancel_all_positions(self):
        assert self.__run_mode != Mode_DryRun
        try:
            if self.position is not None and self.position.q > 0:
                t = SIDE_SELL if self.position.s == SIDE_BUY else SIDE_BUY
                self.client.futures_create_order(symbol=self.symbol, type=FUTURE_ORDER_TYPE_MARKET,
                                                 side=t, quantity=self.position.q)
                self.position = None
        except Exception as e:
            logger.exception(e)

    def cancel_all_orders(self):
        assert self.__run_mode != Mode_DryRun
        for i in range(3):
            # make 3 attempts to cancel until OK
            try:
                res = self.client.futures_cancel_all_open_orders(symbol=self.symbol)
                if res["code"] == 200:
                    return
            except Exception as e:
                logger.exception(e)

    def on_tik(self, realtime_data):
        """Real time data update"""
        return

    def on_start(self):
        logger.info("BOT Start")

    def market_price(self):
        return float(self.client.futures_mark_price(symbol=self.symbol)["markPrice"])

    def balance_usd(self, fetch_from_cache=False):
        assert self.__run_mode != Mode_DryRun
        if fetch_from_cache is False:
            accounts = self.client.futures_account_balance()
            for a in accounts:
                if a["asset"] == "USDT":
                    self.usd = float(a["balance"])
                    return self.usd
            return self.usd
        else:
            return self.usd

    def on_new_interval(self, past_data):
        raise "Unimplemented yet"

    async def reset_data(self):
        dfrom = int(time.time() * 1000) - INTERVAL_2_SECONDS[self.interval] * (self.look_back * 2) * 1000
        # self.data = _get_data_chunk(self.symbol, dfrom, self.interval)[-self.look_back - 1:-1]
        self.data = get_data_all_the_time(self.symbol, dfrom, self.interval, save_path=None)[-self.look_back - 1:-1]

    async def reset_data_util_valid(self, sleep_seconds):
        await self.reset_data()
        while self.validate_data() is False:
            await asyncio.sleep(sleep_seconds)
            await self.reset_data()

    async def loop(self):
        await self.reset_data()
        self.on_start()
        await self.reset_data_util_valid(4)
        logger.info("Data prepare OK")
        self.socket.run_forever()

    def validate_data(self):
        if self.data is None:
            return False
        if len(self.data) != self.look_back:
            return False
        old_t = self.data[-1][0]
        if old_t + INTERVAL_2_SECONDS[self.interval] * 1000 * 1.25 < time.time() * 1000:
            return False
        if old_t + INTERVAL_2_SECONDS[self.interval] * 1000 > time.time() * 1000:
            return False
        if np.count_nonzero(
                np.diff([item[0] for item in self.data]) != INTERVAL_2_SECONDS[self.interval] * 1000):
            return False
        return True

    def on_message(self, ws, message):
        data = ast.literal_eval(message.replace("true", "True").replace("false", "False"))["data"]
        logger.debug(data)
        tdata = data["E"]

        t = time.time() * 1000  # seconds
        if t - tdata < self.data_delay_allowed:
            self.on_tik(self.ws_data_to_kline_data(data['k']))

        if data["k"]["x"] == True:
            self.data.append(self.ws_data_to_kline_data(data['k']))
            self.data = self.data[-self.look_back:]
            if self.validate_data() is False:
                logger.warning("Data invalid, need fix")
                self.reset_data_util_valid(4)
                logger.info("Fixing data done")
        else:
            return

        t = time.time() * 1000  # seconds
        if t - tdata < self.data_delay_allowed:
            logger.info("New interval")
            self.on_new_interval(self.data)

    def on_error(self, wsapp, err):
        logger.error(err)
        if self.stop_on_err:
            self.socket.close()

    async def dry_run(self):
        print("Running in dry mode")
        self.__run_mode = Mode_DryRun
        self.client = Client(testnet=False)
        await self.loop()

    async def run_on_testnet(self, api_key, api_secret):
        self.__run_mode = Mode_Testnet
        self.client = Client(api_key, api_secret, testnet=True)
        await self.loop()

    async def run_live(self, api_key, api_secret):
        self.__run_mode = Mode_Live
        self.client = Client(api_key, api_secret, testnet=False)
        await self.loop()



import numpy as np

from tradeX.datasets.BinanceDataset import BinanceDataset
from tradeX.datasets.consts import SELECTED_FEATURE_IDX


class ClassifierDataset(BinanceDataset):
    EVENT_UP = 1
    EVENT_DOWN = 2
    EVENT_STABLE = 0

    def __init__(self, look_ahead, file, look_back, gap=100, norm=True, limit=-1, feature_engineer=None):
        """
        :param look_ahead:
        :param file:
        :param look_back:
        :param gap: to determine if a candle is up or down. If 0 < abs(openPrice-closePrice) < gap --> STABLE
        :param norm:
        :param limit:
        """
        super(ClassifierDataset, self).__init__(file, look_back=look_back, look_ahead=look_ahead, norm=norm,
                                                limit=limit)
        self.gap = gap
        self.feature_engineer = feature_engineer

    def __getitem__(self, item):
        idx = item + self.look_back
        input = self.data[item:idx]
        gt = self.data[idx:idx + self.look_ahead]

        # take  9 features, except close & open time
        feature_idx = SELECTED_FEATURE_IDX

        input = np.array(input, dtype=np.float32)
        input = input[:, feature_idx]

        # take  9 features, except close & open time

        gt = np.array(gt, dtype=np.float32)
        gt = gt[:, feature_idx]

        event = 0.0 if gt[-1, 3].item() < input[-1, 3].item() else 1.0

        if self.feature_engineer is not None:
            input = self.feature_engineer(input)

        # p_mean = input[:, 0:4].mean()
        # p_std = input[:, 0:4].std()
        mean = input.mean(axis=0)
        std = input.std(axis=0)
        # mean[0:4] = p_mean
        # std[0:4] = p_std
        input = (input - mean) / std
        #gt = (gt - mean) / std

        return {
            "input": input,  # Lb x 9,
            "mean": mean,
            "std": std,
            "ahead": gt,  # La x 9
            "event": event
        }

    def __len__(self):
        return len(self.data) - self.look_back - self.look_ahead + 1



import os
import time
from concurrent.futures import ThreadPoolExecutor
import pandas
import requests
import ast

from pandas import DataFrame
from tacv.fileUtils import save_json, get_all_files, load_json
from loguru import logger


def get_backtest_data(symbol, dfrom, dto, interval, save_path):
    assert dfrom < dto
    if dfrom <= 1638327420000:
        logger.warning("Data before that time was used to train, backtest will not be reliable.")
    dir = os.path.dirname(save_path)
    if not os.path.exists(dir):
        os.makedirs(dir)
    all_data = []
    total_time = dto - dfrom
    chunk = _get_data_chunk(symbol, dfrom=dfrom, interval=interval)
    start_time = time.time() * 1000
    while chunk is not None:
        if len(chunk) > 0 and dfrom < dto:
            dfrom = chunk[-1][6] + 1
            chunk = _get_data_chunk(symbol, dfrom=dfrom, interval=interval)
            all_data.extend(chunk)
            end_time = time.time() * 1000
            remaining_percent = (dto - dfrom) / total_time
            remaining_time = (end_time - start_time) / (1 - remaining_percent) * remaining_percent
            logger.info(f"Estimate to finish in {remaining_time / 1000} seconds.")
        else:
            save_json(save_path, all_data)  # save json
            pdata = DataFrame.from_records(all_data, index="OpenTime",
                                           columns=["OpenTime", "Open", "High", "Low", "Close", "Volume", "CloseTime",
                                                    "QuoteVolume", "NumTrades", "TakerBuyVolume", "TakerBuyQuoteVolume",
                                                    "Ignore"])
            pdata.index = pandas.to_datetime(pdata.index, unit='ms')
            pdata = pdata.drop(columns=["CloseTime", "Ignore"]).astype(float)
            DataFrame.to_csv(pdata, save_path.replace("json", "csv"), index=True)
            return pdata
    else:
        print(f"Error when getting data, params: symbol={symbol}, interval={interval},dfrom={dfrom}")

    return all_data


def get_data_all_the_time(symbol, dfrom, interval, save_path):
    if save_path is not None:
        base_name = os.path.basename(save_path)
        dir = os.path.dirname(save_path)
    all_data = []
    chunk = _get_data_chunk(symbol, dfrom=dfrom, interval=interval)
    i = 0
    while chunk is not None:
        if len(chunk) > 0:
            dfrom = chunk[-1][6] + 1
            chunk = _get_data_chunk(symbol, dfrom=dfrom, interval=interval)
            all_data.extend(chunk)
        else:
            # merge data_temp files
            if save_path is not None:
                all_files = get_all_files(dir, recursive=False, exts=["json"])
                all_files = [f for f in all_files if "temp_" in f]
                all_files.sort()
                all_data = []
                for f in all_files:
                    all_data.extend(load_json(f))
                save_json(save_path, all_data)
                print(f"Removing {len(all_files)} file(s).")
                for f in all_files:
                    os.remove(f)
                logger.info("Done getting all data.")
            break

        if i % 10 == 0:
            if save_path is not None:
                logger.info(f"Saving temp data ({len(all_data)} rows)")
                save_json(os.path.join(dir, "temp_" + f"{i}".zfill(10) + base_name), all_data)
                all_data = []
        i += 1
    else:
        print(f"Error when getting data, params: symbol={symbol}, interval={interval},dfrom={dfrom}")
    return all_data


def _get_data_chunk(symbol, dfrom, interval, save_to=None):
    """
    :param symbol: .e.g. BTCUSDT
    :param dfrom: milliseconds since epoch
    :param dto: milliseconds since epoch
    :param interval: 1m, 1h, 1d,...
    :return: json array of trade
    """
    endpoint = f"https://fapi.binance.com/fapi/v1/klines?symbol={symbol}&interval={interval}&startTime={dfrom}"
    res = requests.get(endpoint)
    if res.status_code != 200:
        print(f"Error, reason {res.reason}, content {res.content}")
        return None
    else:
        data = ast.literal_eval(res.text)
        if save_to is not None:
            save_json(save_to, data)
        return data



import numpy as np
import talib


def remove_nan(ta_signal: np.array):
    """
    :param ta_signal: K x F
    :return:
    """
    max_nan_index = 0
    max_inf_index = 0
    nan_ids = np.nonzero(np.isnan(ta_signal))[0]
    if len(nan_ids) > 0:
        max_nan_index = nan_ids.max() + 1
    inf_ids = np.nonzero(np.isinf(ta_signal))[0]
    if len(inf_ids) > 0:
        max_inf_index = inf_ids.max() + 1

    cut_idx = max(max_nan_index, max_inf_index)
    return ta_signal[cut_idx:].astype(np.float32)


def feature_engineer_fake(input: np.ndarray):
    """
    :param input: L x 9 where 9 is of ("Open", "High", "Low", "Close", "Vol", "Quote asset volume",
             "Number of trades", "Taker buy base asset volume", "Taker buy quote asset volume")
    :param out_seq_length:
    :return:
    """
    """Do nothing, just return"""
    return input


def feature_engineer_20(inp: np.ndarray):
    """("Open", "High", "Low", "Close", "Vol", "Quote asset volume",
             "Number of trades", "Taker buy base asset volume", "Taker buy quote asset volume")"""

    inp = inp.astype(float)
    open = inp[:, 0]
    high = inp[:, 1]
    low = inp[:, 2]
    close = inp[:, 3]
    vol = inp[:, 4]
    quote_asset_vol = inp[:, 5]
    num_trades = inp[:, 6]
    tbbav = inp[:, 7]
    tbqav = inp[:, 8]
    sma = talib.SMA(inp[:, 3], timeperiod=10)
    ema = talib.EMA(inp[:, 3], timeperiod=10)
    rsi = talib.RSI(inp[:, 3], timeperiod=14)
    macd, macd_signal, macd_hist = talib.MACD(inp[:, 3], fastperiod=12, slowperiod=26, signalperiod=9)
    stoch_k, stoch_d = talib.STOCH(inp[:, 1], inp[:, 2], inp[:, 3], fastk_period=14, slowk_period=3, slowd_period=3)
    bbands_upper, bbands_middle, bbands_lower = talib.BBANDS(inp[:, 3], timeperiod=20, nbdevup=2, nbdevdn=2, matype=0)
    atr = talib.ATR(inp[:, 1], inp[:, 2], inp[:, 3], timeperiod=14)
    cci = talib.CCI(inp[:, 1], inp[:, 2], inp[:, 3], timeperiod=20)
    obv = talib.OBV(inp[:, 3], inp[:, 4])
    willr = talib.WILLR(inp[:, 1], inp[:, 2], inp[:, 3], timeperiod=14)
    mfi = talib.MFI(inp[:, 1], inp[:, 2], inp[:, 3], inp[:, 4], timeperiod=14)
    roc = talib.ROC(inp[:, 3], timeperiod=10)
    adx = talib.ADX(inp[:, 1], inp[:, 2], inp[:, 3], timeperiod=14)
    trix = talib.TRIX(inp[:, 3], timeperiod=10)
    uo = talib.ULTOSC(inp[:, 1], inp[:, 2], inp[:, 3], timeperiod1=7, timeperiod2=14, timeperiod3=28)
    signals = [open, high, low, close, vol, quote_asset_vol, num_trades, tbbav, tbqav, sma, ema, rsi, macd, macd_signal,
               macd_hist, stoch_k, stoch_d, bbands_upper,
               bbands_middle, bbands_lower, atr, cci, obv, willr, mfi, roc, adx, trix, uo]
    f = np.vstack(signals).T

    # remember to always keep this line to remove NaN values which usually appears in head of each signal
    f = remove_nan(f)
    return f


def feature_engineer_20_half_time_period(inp: np.ndarray):
    """("Open", "High", "Low", "Close", "Vol", "Quote asset volume",
             "Number of trades", "Taker buy base asset volume", "Taker buy quote asset volume")"""

    inp = inp.astype(float)
    open = inp[:, 0]
    high = inp[:, 1]
    low = inp[:, 2]
    close = inp[:, 3]
    vol = inp[:, 4]
    quote_asset_vol = inp[:, 5]
    num_trades = inp[:, 6]
    tbbav = inp[:, 7]
    tbqav = inp[:, 8]
    sma = talib.SMA(inp[:, 3], timeperiod=5)
    ema = talib.EMA(inp[:, 3], timeperiod=5)
    rsi = talib.RSI(inp[:, 3], timeperiod=7)
    macd, macd_signal, macd_hist = talib.MACD(inp[:, 3], fastperiod=6, slowperiod=13, signalperiod=5)
    stoch_k, stoch_d = talib.STOCH(inp[:, 1], inp[:, 2], inp[:, 3], fastk_period=7, slowk_period=3, slowd_period=3)
    bbands_upper, bbands_middle, bbands_lower = talib.BBANDS(inp[:, 3], timeperiod=10, nbdevup=2, nbdevdn=2, matype=0)
    atr = talib.ATR(inp[:, 1], inp[:, 2], inp[:, 3], timeperiod=7)
    cci = talib.CCI(inp[:, 1], inp[:, 2], inp[:, 3], timeperiod=10)
    obv = talib.OBV(inp[:, 3], inp[:, 4])
    willr = talib.WILLR(inp[:, 1], inp[:, 2], inp[:, 3], timeperiod=7)
    mfi = talib.MFI(inp[:, 1], inp[:, 2], inp[:, 3], inp[:, 4], timeperiod=7)
    roc = talib.ROC(inp[:, 3], timeperiod=5)
    adx = talib.ADX(inp[:, 1], inp[:, 2], inp[:, 3], timeperiod=7)
    trix = talib.TRIX(inp[:, 3], timeperiod=5)
    uo = talib.ULTOSC(inp[:, 1], inp[:, 2], inp[:, 3], timeperiod1=4, timeperiod2=7, timeperiod3=14)
    signals = [open, high, low, close, vol, quote_asset_vol, num_trades, tbbav, tbqav, sma, ema, rsi, macd, macd_signal,
               macd_hist, stoch_k, stoch_d, bbands_upper,
               bbands_middle, bbands_lower, atr, cci, obv, willr, mfi, roc, adx, trix, uo]
    f = np.vstack(signals).T

    # remember to always keep this line to remove NaN values which usually appears in head of each signal
    f = remove_nan(f)
    return f


def feature_engineer_example(input: np.ndarray):
    """
    :param input: L x 9 where 9 is of ("Open", "High", "Low", "Close", "Vol", "Quote asset volume",
             "Number of trades", "Taker buy base asset volume", "Taker buy quote asset volume")
    :param out_seq_length:
    :return:
    """
    # always add this line to convert input to float64
    input = input.astype(float)

    # calculate TA signals
    open = input[:, 0]
    high = input[:, 1]
    low = input[:, 2]
    close = input[:, 3]
    vol = input[:, 4]

    signals = [open, high, low, close]
    signals.append(talib.SMA(close, timeperiod=4))
    signals.append(talib.EMA(close, timeperiod=4))
    signals.append(talib.PLUS_DI(high, low, close, timeperiod=4))

    signals.append(talib.RSI(close, timeperiod=4))
    signals.append(talib.STOCHRSI(close, timeperiod=4))
    signals.append(talib.MA(close, timeperiod=4))
    signals.append(talib.BBANDS(close, timeperiod=4))
    signals.append(talib.MACD(close))
    signals.append(talib.OBV(close, vol))
    signals.append(talib.ADX(high, low, close, timeperiod=4))
    signals.append(talib)
    f = np.vstack(signals).T

    # remember to always keep this line to remove NaN values which usually appears in head of each signal
    f = remove_nan(f)
    return f


def feature_engineer_2(input: np.ndarray):
    """
    :param input: L x 9 where 9 is of ("Open", "High", "Low", "Close", "Vol", "Quote asset volume",
             "Number of trades", "Taker buy base asset volume", "Taker buy quote asset volume")
    :param out_seq_length:
    :return:
    """
    # always add this line to convert input to float64
    input = input.astype(float)

    # calculate TA signals
    high = input[:, 1]
    low = input[:, 2]
    close = input[:, 3]
    vol = input[:, 4]
    signals = []
    signals.append(talib.PLUS_DI(high, low, close, timeperiod=4))
    signals.append(talib.EMA(close, timeperiod=4))
    signals.append(talib.RSI(close, timeperiod=4))
    signals.append(talib.STOCHRSI(close, timeperiod=4))
    signals.append(talib.MA(close, timeperiod=4))
    signals.append(talib.BBANDS(close, timeperiod=4))
    signals.append(talib.MACD(close))
    signals.append(talib.OBV(close, vol))
    signals.append(talib.ADX(high, low, close, timeperiod=4))
    f = np.vstack(signals).T

    f = remove_nan(f)
    mean = f.mean(axis=0)
    std = f.std(axis=0)
    f = (f - mean) / std
    return f



from collections import defaultdict

import numpy as np


def far_at_k_hits(true_labels: np.ndarray, scores: np.ndarray, k_hits: float = 1000):
    """
    Finds the threshold for positive samples that corresponds to a specified FAR for each class in a multi-class
    classification problem, given the true labels and confidence scores.

    Parameters:
        true_labels (numpy.ndarray): An array of shape (n_samples,) containing the true labels for each sample.
        scores (numpy.ndarray): An array of shape (n_samples, n_classes) containing the confidence scores for each class
                                for each sample.
        far (float): The desired false alarm rate (FAR) for each class.

    Returns:
        thresholds (numpy.ndarray): An array of shape (n_classes,) containing the threshold that corresponds to the
                                     desired FAR for each class.
    """

    # Convert the true labels to one-hot encoding
    num_classes = scores.shape[1]

    # Initialize variables to keep track of the number of true negatives and false positives
    all_class_hits = defaultdict(list)
    all_class_far = defaultdict(list)
    all_threshold = defaultdict(list)
    for clz_id in range(num_classes):
        pred_scores = scores[:, clz_id]
        sorted_idx = pred_scores.argsort()
        sorted_score = pred_scores[sorted_idx]
        gts = true_labels[sorted_idx]
        for i in range(len(gts)):
            threshold = sorted_score[i]
            TP = np.count_nonzero(gts[i:] == clz_id)
            FP = np.count_nonzero(gts[i:] != clz_id)
            all_class_hits[clz_id].append(len(gts[i:]))
            all_class_far[clz_id].append(FP / ((len(gts[i:]) + 0.0000001)))
            all_threshold[clz_id].append(threshold)
    c = 1
    far, threshold = np.concatenate((all_class_hits[c], all_class_far[c], all_threshold[c])).reshape(3, -1)[1:, -1000]
    return far, threshold



from binance.enums import SIDE_BUY, SIDE_SELL
from torch.utils.tensorboard import SummaryWriter

from tradeX.bots.binance_bot import BinanceBot
from tradeX.models.EModel import EModelClassifier

import numpy as np


class AbnormalDetectionBot(BinanceBot):
    """
           This bot is to measure model performance on live market
           """

    def __init__(self, model: EModelClassifier, logger: SummaryWriter, *args, **kargs):
        super(AbnormalDetectionBot, self).__init__(*args, **kargs)
        self.model = model
        self.logger = logger
        self.leverage = 10
        self.iter = 0
        self.track = {"position": self.position, "iter": self.iter, "usd_at_entry": self.usd}
        self.lose = 0
        self.win = 0

    def open_track(self, current_price, side):
        q = round(self.usd * self.leverage / current_price, 3)
        fee = q * current_price * 0.04 / 100
        self.track["usd_at_entry"] = self.usd
        self.usd -= fee
        self.track["position"] = BinanceBot.Position(q=q, s=side, entry=current_price)
        self.track["iter"] = self.iter

    def close_track(self, current_price):
        t = self.track
        p = self.track["position"]
        if p is not None:
            q = p.q
            s = p.s
            fee = q * current_price * 0.04 / 100
            profit = q * (current_price - p.entry) * (1 if s == SIDE_BUY else -1)
            before = t["usd_at_entry"]
            self.usd = before - fee + profit
            self.track["position"] = None
            self.track["iter"] = 0
            if self.usd <= before:
                self.lose += 1
            else:
                self.win += 1

    def on_start(self):
        print("Start==============================================")

    def on_tik(self, realtime_data):
        data = np.array(realtime_data).astype(float).tolist()
        _, o, h, l, c, *_ = data  # c = current price
        t = self.track
        p = t["position"]
        estimate_usd = self.usd
        if p is not None:
            estimate_usd += (c - p.entry) * p.q * (1 if p.s == SIDE_BUY else -1)
        self.logger.add_scalar("USD", estimate_usd, self.iter)

    def on_new_interval(self, pdata):
        """
        :param pdata: array of size self.look_back x 12
        :return:
        """
        self.iter += 1
        self.logger.add_scalar("Performance/Win", self.win, self.iter)
        self.logger.add_scalar("Performance/Lose", self.lose, self.iter)

        current_price = float(pdata[-1][4])

        if self.track["position"] is not None and self.track["iter"] >= 15:
            self.close_track(current_price)

        bump, dump = self.model.predict(pdata)
        if bump > dump > 0.1:
            if self.track["position"] is not None:
                if self.track["position"].s != SIDE_BUY:
                    self.close_track(current_price)
                    self.open_track(current_price, side=SIDE_BUY)
            else:
                self.open_track(current_price, side=SIDE_BUY)

        elif dump > bump > 0.1:
            if self.track["position"] is not None:
                if self.track["position"].s != SIDE_SELL:
                    self.close_track(current_price)
                    self.open_track(current_price, side=SIDE_SELL)
            else:
                self.open_track(current_price, side=SIDE_SELL)



import asyncio

import plotly.graph_objects as go
import streamlit as st
import numpy as np
from tradeX.bots import BinanceDataBot


class StreamListView(BinanceDataBot):
    def __init__(self, *args, **kwargs):
        super(StreamListView, self).__init__(*args, **kwargs)

    async def render(self):
        st.title("TradeX")
        await self.loop()

    def on_new_interval(self, past_data):
        # process data
        # make prediction
        prediction = self.predict(np.array(past_data))
        # plot
        self.plot_prediction_on_chart(np.array(past_data), prediction)

    def on_start(self):
        prediction = self.predict(self.data)
        # plot
        self.plot_prediction_on_chart(np.array(self.data), prediction)

    def plot_prediction_on_chart(self, raw_data, prediction):
        fig = go.Figure()
        fig.add_trace(go.Candlestick(x=raw_data[:, 0], open=raw_data[:, 1], high=raw_data[:, 2], low=raw_data[:, 3],
                                     close=raw_data[:, 4], name="candle_stick"))
        # Set the chart layout
        fig.update_layout(title="Bitcoin Price",
                          xaxis_title="Date",
                          yaxis_title="Price",
                          hovermode="x unified")

        # Set the Plotly layout options, including scrollZoom and dragmode
        layout = go.Layout(uirevision="foo", dragmode="pan", hovermode="closest")

        # Create a FigureWidget from the chart and layout
        fig_widget = go.FigureWidget(fig, layout=layout)

        # Render the chart using Streamlit
        st.plotly_chart(fig_widget)

    def predict(self, data):
        return 1


if __name__ == '__main__':
    view = StreamListView(interval="1m", look_back=1000, symbol="btcusdt")
    asyncio.run(view.render())



from binance.enums import SIDE_BUY, SIDE_SELL
from torch.utils.tensorboard import SummaryWriter

from tradeX.bots.binance_bot import BinanceBot
from tradeX.models.EModel import EModelClassifier

import numpy as np


class LiveAbnormalDetectionBot(BinanceBot):
    """
           This bot is to measure model performance on live market
           """

    def __init__(self, model: EModelClassifier, logger: SummaryWriter, *args, **kargs):
        super(LiveAbnormalDetectionBot, self).__init__(*args, **kargs)
        self.model = model
        self.logger = logger
        self.leverage = 10
        self.iter = 0
        self.lose = 0
        self.win = 0

    def on_start(self):
        print("Start==============================================")

    def on_tik(self, realtime_data):
        data = np.array(realtime_data).astype(float).tolist()
        _, o, h, l, c, *_ = data  # c = current price
        t = self.track
        p = t["position"]
        estimate_usd = self.usd
        if p is not None:
            estimate_usd += (c - p.entry) * p.q * (1 if p.s == SIDE_BUY else -1)
        self.logger.add_scalar("USD", estimate_usd, self.iter)

    def on_new_interval(self, pdata):
        """
        :param pdata: array of size self.look_back x 12
        :return:
        """
        self.iter += 1
        # self.logger.add_scalar("Performance/Win", self.win, self.iter)
        # self.logger.add_scalar("Performance/Lose", self.lose, self.iter)
        self.logger.add_scalar("USD", self.balance_usd())

        bump, dump = self.model.predict(pdata)
        print(bump, dump)
        if bump > dump > 0.11:
            if self.position is not None:
                if self.position.s != SIDE_BUY:
                    self.cancel_all_positions()
                    self.long_market(0.9, self.leverage)
            else:
                self.long_market(0.9, self.leverage)

        elif dump > bump > 0.11:
            if self.position is not None:
                if self.position.s != SIDE_SELL:
                    self.cancel_all_positions()
                    self.short_market(0.9, self.leverage)
            else:
                self.short_market(0.9, self.leverage)



import math
from torch import tanh

import mlflow
import torch
from pytorch_lightning import LightningModule
from torch import Tensor
from torch.nn import Linear, TransformerEncoderLayer, LayerNorm, TransformerEncoder, \
    TransformerDecoderLayer, TransformerDecoder, Module, Dropout
from torch.nn.functional import l1_loss, max_pool1d, relu

from tradeX.utils.losses import log_return_loss


class ITTransformer(LightningModule):
    def __init__(self, d_model: int = 512, nhead: int = 1, dim_feedforward: int = 2048,
                 output_d=9, input_d=9, d_in_decoder=4, position_encode=False):
        super(ITTransformer, self).__init__()

        # modules
        self.fc_in_en = Linear(input_d, d_model)
        self.pos_encoder = PositionalEncoder(max_seq_len=1000, d_model=d_model)

        encoder_layer = TransformerEncoderLayer(d_model, nhead, dim_feedforward, batch_first=True)
        encoder_norm = LayerNorm(d_model)
        self.encoder = TransformerEncoder(encoder_layer, 6, encoder_norm)

        # self.fc_in_de = Linear(4, d_model)
        # decoder_layer = TransformerDecoderLayer(d_model, nhead, dim_feedforward, batch_first=True)
        # decoder_norm = LayerNorm(d_model)
        # self.decoder = TransformerDecoder(decoder_layer, 6, decoder_norm)

        self.fc_reg = Linear(d_model, output_d)

        self.position_encode = position_encode
        self.init_weights()

    def init_weights(self) -> None:
        initrange = 0.1
        for layer in [self.fc_in_en, self.fc_reg]:
            layer.weight.data.uniform_(-initrange, initrange)
            layer.bias.data.zero_()
            layer.weight.data.uniform_(-initrange, initrange)

    def forward(self, src):
        """
        :param src: B x L x D_in
        :param tgt: B x W x D_out
        :param tgt_y: same as $tgt
        :return: shifted right of $tgt in shape of B x W x D_out
        """
        target = src[:, -1:, :4]
        x = self.fc_in_en(src)  # B x L x D
        x = relu(x)
        if self.position_encode:
            x = self.pos_encoder(x)
        mem = self.encoder(x)  # B x L x D
        mem = relu(mem)  # B x L x D
        out = self.fc_reg(mem[:,-1]) # B x D_out
        return out
        # decoder_in = self.fc_in_de(target)  # B x 1 x D
        # decoder_in = relu(decoder_in)
        # out = self.decoder(decoder_in, mem)  # B x 1 x D
        # out = relu(out)
        # out = self.fc_reg(out)  # B x 1 x D_out
        # return out.view(len(out), -1)

    def train_val_step(self, batch, batch_id, key_log):
        input = batch["input"].float()  # B x L_back x D_in
        gt_reg = batch["log_return"].float()  # B x L_ahead x D_in
        log_return_pred = self(input)  # B x 3

        loss = l1_loss(log_return_pred, gt_reg, reduction="mean")

        self.log(f"{key_log}_loss", loss.item())
        mlflow.log_metric(f"{key_log}_loss", loss.item())
        return loss

    def training_step(self, batch, batch_id):
        self.log("learning_rate", self.lr_scheduler.get_last_lr()[0])
        return self.train_val_step(batch, batch_id, "train")

    def validation_step(self, batch, batch_id):
        return self.train_val_step(batch, batch_id, "val")

    def set_optimizer(self, optimizer, lr_scheduler):
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler

    def configure_optimizers(self):
        return [self.optimizer], [self.lr_scheduler]


class PositionalEncoder(Module):
    """
    The authors of the original transformer paper describe very succinctly what
    the positional encoding layer does and why it is needed:

    "Since our model contains no recurrence and no convolution, in order for the
    model to make use of the order of the sequence, we must inject some
    information about the relative or absolute position of the tokens in the
    sequence." (Vaswani et al, 2017)
    Adapted from:
    https://pytorch.org/tutorials/beginner/transformer_tutorial.html
    """

    def __init__(
            self,
            dropout: float = 0.1,
            max_seq_len: int = 5000,
            d_model: int = 512,
            batch_first: bool = False
    ):
        """
        Parameters:
            dropout: the dropout rate
            max_seq_len: the maximum length of the input sequences
            d_model: The dimension of the output of sub-layers in the model
                     (Vaswani et al, 2017)
        """

        super().__init__()

        self.d_model = d_model

        self.dropout = Dropout(p=dropout)

        self.batch_first = batch_first

        self.x_dim = 1 if batch_first else 0

        # copy pasted from PyTorch tutorial
        position = torch.arange(max_seq_len).unsqueeze(1)

        div_term = torch.exp(torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model))

        pe = torch.zeros(max_seq_len, 1, d_model)

        pe[:, 0, 0::2] = torch.sin(position * div_term)

        pe[:, 0, 1::2] = torch.cos(position * div_term)

        self.register_buffer('pe', pe)

    def forward(self, x: Tensor) -> Tensor:
        """
        Args:
            x: Tensor, shape [batch_size, enc_seq_len, dim_val] or
               [enc_seq_len, batch_size, dim_val]
        """

        x = x + self.pe[:x.size(self.x_dim)]

        return self.dropout(x)



import setuptools
from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
with open("requirements.txt", "r") as f:
    lines = f.readlines()
    required_pkgs = [item.strip() for item in lines]

new_ver = "0.0.3"
setup(
    name='tradeX',
    version=new_ver,
    packages=setuptools.find_packages(where="."),
    license='LICENSE',
    author='TaQuangTu',
    install_requires=required_pkgs,
    author_email='taquangtu132@gmail.com',
    description='Machine learning based crypto currency price prediction',
    long_description_content_type="text/markdown",
    long_description=long_description,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import time
from typing import Union

import numpy as np
import torch

from tradeX.datasets.consts import SELECTED_FEATURE_IDX
from tradeX.utils.data_getter import _get_data_chunk


class DataNormalizer:
    def norm(self, data, *args, **kwargs):
        pass

    def un_norm(self, data, **kwargs):
        pass


class GaussianNormalizer(DataNormalizer):
    def norm(self, data: Union[np.ndarray, torch.Tensor], mean=None, std=None):
        if mean is None or std is None:
            mean = data.mean()
            std = data.std()
        data = (data - mean) / std
        return data, {"mean": mean, "std": std}

    def un_norm(self, data, **kwargs):
        assert "mean" in kwargs and "std" in kwargs, "kwargs must contains mean & std value to decode"
        mean = kwargs["mean"]
        std = kwargs["std"]
        data = data * std + mean
        return data


class MinMaxNormalizer(DataNormalizer):
    def norm(self, data: Union[np.ndarray, torch.Tensor], min=None, max=None):
        if min is None or max is None:
            max = data.max()
            min = data.min()
        data = (data - min) / (max - min)
        return data, {"min": min, "max": max}

    def un_norm(self, data, **kwargs):
        assert "max" in kwargs and "min" in kwargs, "kwargs must contains max min value to decode"
        min = kwargs["min"]
        max = kwargs["max"]
        data = (data + min) * (max - min)
        return data


# def norm_input(input):
#     mms = []
#     normalizer = GaussianNormalizer()
#     for i in range(input.shape[1]):
#         input[:, i], mm = normalizer.norm(input[:, i])
#         mms.append(mm)
#     return input, mms


def un_norm_output(input, mms):
    for i in range(len(input)):
        mean = mms[i]["mean"]
        std = mms[i]["std"]
        input[i] = input[i] * std + mean
    return input


def download_pastdata(symbol, interval, dfrom, save_to):
    data = _get_data_chunk(symbol, dfrom=dfrom, interval=interval, save_to=save_to)
    return data


def download_livedata(symbol, interval):
    data = _get_data_chunk(symbol, dfrom=time.time() * 1000, interval=interval)
    return data


def norm_data(input, gt, norm_gt=True):
    """
    :param input: L x 9
    :param gt: L x 9
    :param norm_gt:
    :return:
    """
    normalizer = GaussianNormalizer()
    ms = []

    price_mean = input[:, 0:4].mean()
    price_std = input[:, 0:4].std()
    for i in range(input.shape[1]):
        if 0 <= i <= 3:
            input[:, i], mm = normalizer.norm(input[:, i], mean=price_mean, std=price_std)
        else:
            input[:, i], mm = normalizer.norm(input[:, i])
        if norm_gt is True:
            gt[:, i], _ = normalizer.norm(gt[:, i], **mm)
        ms.append(mm)
    return input, gt, ms


def process_input(data, window_size, to_torch=True, do_norm=True):
    input = np.array(data, dtype=np.float32)

    # take  9 features, except close & open time
    feature_idx = SELECTED_FEATURE_IDX
    input = input[:, feature_idx]
    mms = None
    if do_norm:
        input, _, mms = norm_data(input, None, norm_gt=False)
    if to_torch:
        input = torch.from_numpy(input).view(window_size, -1).float()
    else:
        # keep numpy
        input = input.reshape(window_size, -1)
    return input, mms


def feature_engineering_volume(in_data: np.ndarray, is_normalized: bool = False):
    """
    :param is_normalized: indicate if the input data is already normalized or not
    :param in_data: L x K for non-batch input or B x L x K for batch input
    :return:  L x K1 for non-batch input or B x L x K1 for batch input
    Where K and K1 are number of features before and after feature engineering
    K = 9
    """
    if is_normalized:
        raise "Feature engineering is now only supports non-normalized data"
    # [open, high, low, close, volume, qv, n, tbbav, tbqav]

    v_and_qv = in_data[..., 4:6]  #
    tbba_tbqa = in_data[..., 7:9]

    unnormed_features = np.concatenate((v_and_qv, tbba_tbqa),
                                       axis=-1)

    mean = unnormed_features.mean()
    std = unnormed_features.std()
    normed_features = (unnormed_features - mean) / std

    return normed_features



import ast
import asyncio
import time

import numpy as np
from loguru import logger
from websocket import WebSocketApp

from tradeX.bots.consts import INTERVAL_2_SECONDS
from tradeX.utils.data_getter import get_data_all_the_time


class BinanceDataBot:
    def __init__(self, symbol, interval, look_back=100, data_delay_allowed=5000, stop_on_err=True):
        self.symbol = symbol
        self.interval = interval
        self.look_back = look_back
        self.data_delay_allowed = data_delay_allowed
        self.data = []
        self.stop_on_err = stop_on_err
        self.socket = WebSocketApp(
            f"wss://fstream.binance.com/stream?streams={self.symbol}@kline_{self.interval}",
            on_message=self.on_message,
            on_error=self.on_error)

    @staticmethod
    def ws_data_to_kline_data(ws_data):
        open_t = ws_data['t']
        open = ws_data['o']
        high = ws_data['h']
        low = ws_data['l']
        close = ws_data['c']
        volume = ws_data['v']
        close_t = ws_data['T']
        qv = ws_data['q']
        n = ws_data['n']
        tbbav = ws_data['V']
        tbqav = ws_data['Q']
        ignore = ws_data['B']
        return [open_t, open, high, low, close, volume, close_t, qv, n, tbbav, tbqav, ignore]

    def on_tik(self, realtime_data):
        """Real time data update"""
        return

    def on_start(self):
        logger.info("BOT Start")

    def on_new_interval(self, past_data):
        raise "Unimplemented yet"

    async def reset_data(self):
        dfrom = int(time.time() * 1000) - INTERVAL_2_SECONDS[self.interval] * (self.look_back * 2) * 1000
        self.data = get_data_all_the_time(self.symbol, dfrom, self.interval, save_path=None)[-self.look_back - 1:-1]

    async def reset_data_util_valid(self, sleep_seconds):
        await self.reset_data()
        while self.validate_data() is False:
            await asyncio.sleep(sleep_seconds)
            await self.reset_data()

    async def loop(self):
        await self.reset_data()
        self.on_start()
        await self.reset_data_util_valid(4)
        logger.info("Data prepare OK")
        self.socket.run_forever()

    def validate_data(self):
        if self.data is None:
            return False
        if len(self.data) != self.look_back:
            return False
        old_t = self.data[-1][0]
        if old_t + INTERVAL_2_SECONDS[self.interval] * 1000 * 1.25 < time.time() * 1000:
            return False
        if old_t + INTERVAL_2_SECONDS[self.interval] * 1000 > time.time() * 1000:
            return False
        if np.count_nonzero(
                np.diff([item[0] for item in self.data]) != INTERVAL_2_SECONDS[self.interval] * 1000):
            return False
        return True

    def on_message(self, ws, message):
        data = ast.literal_eval(message.replace("true", "True").replace("false", "False"))["data"]
        logger.debug(data)
        tdata = data["E"]

        t = time.time() * 1000  # seconds
        if t - tdata < self.data_delay_allowed:
            self.on_tik(self.ws_data_to_kline_data(data['k']))

        if data["k"]["x"] == True:
            self.data.append(self.ws_data_to_kline_data(data['k']))
            self.data = self.data[-self.look_back:]
            if self.validate_data() is False:
                logger.warning("Data invalid, need fix")
                self.reset_data_util_valid(4)
                logger.info("Fixing data done")
        else:
            return

        t = time.time() * 1000  # seconds
        if t - tdata < self.data_delay_allowed:
            logger.info("New interval")
            self.on_new_interval(self.data)

    def on_error(self, wsapp, err):
        logger.error(err)
        if self.stop_on_err:
            self.socket.close()






class EModel:
    def __init__(self, model):
        self.model = model

    def load_model(self, ckpt_path):
        raise "Not implement yet"

    def predict(self, input, *args, **kargs):
        """
        :param input: The last 100-interval raw trading data
        :return: predictions for High price, Low price, Close price for next interval based on history data
        """
        data, mms = process_input(input, len(input))

        if "ahead" in kargs:
            ahead = kargs["ahead"]
        else:
            ahead = 1

        # predict and unnorm predictions
        predict = self.model(data, ahead=ahead).view(3)

        predict = un_norm_output(predict, mms[1:]).tolist()

        # pred high, low, close price
        php, plp, pcp = predict[:3]
        return php, plp, pcp



import numpy as np
from tacv.fileUtils import load_json
from torch.utils.data import Dataset

from tradeX.datasets.consts import SELECTED_FEATURE_IDX
from tradeX.utils.data import norm_data


class BinanceDataset(Dataset):

    def __init__(self, file: str, look_back: int, look_ahead: int = 1, norm: bool = True, limit: int = -1):
        self.look_back = look_back
        self.look_ahead = look_ahead
        self.data = load_json(file)
        if limit > 0:
            self.data = self.data[:limit]
        self.norm = norm

    def __getitem__(self, item):
        inp = self.data[item:item + self.look_back]
        gt = self.data[item + self.look_back:item + self.look_back + self.look_ahead]

        # convert to numpy
        inp = np.array(inp)
        gt = np.array(gt)

        # take 9 features, except close & open time
        feature_idx = SELECTED_FEATURE_IDX
        inp = inp[:, feature_idx]  # L_back x 9
        gt = gt[:, feature_idx]  # L_ahead x 9

        mm = 0
        if self.norm:
            inp, gt, mm = norm_data(inp, gt, norm_gt=True)

        # input : L_back x 9
        # gt : L_ahead x 9
        return {
            "input": inp,  # L_back x 9
            "gt": gt,  # L_ahead x 9
            "mm": mm if self.norm else 0,  # 9
        }

    def __len__(self):
        return len(self.data) - self.look_back - self.look_ahead + 1
